/*
 * MacSamba
 * <http://ghostwood.org/software/macSamba.html>
 * Copyright 2006 Nick Howes <nick@ghostwood.org>
 * tabs every 4 spaces
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2
 * of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <mach/mach.h>
#include <CoreFoundation/CoreFoundation.h>
#include <IOKit/IOKitLib.h>
#include <IOKit/IOCFPlugIn.h>
#include <IOKit/usb/IOUSBLib.h>

#include "util.h"
#include "usb.h"



IOUSBDeviceInterface** findDevice(mach_port_t masterPort, SInt32 vendor_id, SInt32 product_id, USBDeviceAddress usb_address, UInt32 location_id) {
	kern_return_t err;
	CFMutableDictionaryRef matchingDict = 0;
	io_iterator_t iterator = 0;
	io_service_t usbDevice = 0;
	IOCFPlugInInterface **plugInInterface = 0;
	IOUSBDeviceInterface **device = 0;
	SInt32 score;
	HRESULT result = 0;
	USBDeviceAddress usbAddress = 0;
	UInt32 locationID = 0;
	
	matchingDict = IOServiceMatching(kIOUSBDeviceClassName);
	if (matchingDict == 0) failExit("IOServiceMatching(kIOUSBDeviceClassName)", (int)err);

	CFDictionarySetValue(matchingDict, CFSTR(kUSBVendorID), CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &vendor_id));
	CFDictionarySetValue(matchingDict, CFSTR(kUSBProductID), CFNumberCreate(kCFAllocatorDefault, kCFNumberSInt32Type, &product_id));
	err = IOServiceGetMatchingServices(masterPort, matchingDict, &iterator);
	matchingDict = 0;
	if (err != KERN_SUCCESS) failExit("IOServiceGetMatchingServices", (int)err);
	
	while ((usbDevice = IOIteratorNext(iterator)) != 0) {
		err = KERN_SUCCESS;
		
		err = IOCreatePlugInInterfaceForService(usbDevice, kIOUSBDeviceUserClientTypeID, kIOCFPlugInInterfaceID, &plugInInterface, &score);
		if (err != KERN_SUCCESS) failExit("IOCreatePlugInInterfaceForService", (int)err);
		
		result = (*plugInInterface)->QueryInterface(plugInInterface, CFUUIDGetUUIDBytes (kIOUSBDeviceInterfaceID), (LPVOID)&device);
		if (result) failExit("QueryInterface", (int)result);

		(*plugInInterface)->Release(plugInInterface);
		plugInInterface = 0;
		
		(*device)->GetDeviceAddress(device, &usbAddress);
		(*device)->GetLocationID(device, &locationID);
		
		if (usb_address == 0 && location_id == 0) break;
		if (location_id == locationID) break;
		if (usb_address == usbAddress) break;
		(*device)->Release (device);
		device = 0;
	}
	
	if (device == 0) failExit("USB device not found", 0);
	
	if (usbDevice) IOObjectRelease(usbDevice);
	if (iterator) IOObjectRelease(iterator);
	
	return (device);
}


void setConfiguration(IOUSBDeviceInterface **device, int configId) {
	kern_return_t err;
	
	err = (*device)->USBDeviceOpen(device);
	if (err != KERN_SUCCESS) failExit("USBDeviceOpen", (int)err);
	
	err = (*device)->ResetDevice(device);
	if (err != KERN_SUCCESS) failExit("ResetDevice", (int)err);
	
	err = (*device)->SetConfiguration(device, configId);
	if (err != KERN_SUCCESS) failExit("SetConfiguration", (int)err);
	
	(*device)->USBDeviceClose(device);
}


IOUSBInterfaceInterface190** findInterface(IOUSBDeviceInterface **device) {
	kern_return_t err;
	UInt8 configuration_id;
	IOUSBFindInterfaceRequest request;
	io_iterator_t iterator = 0;
	io_service_t usbInterface = 0;
	IOCFPlugInInterface **plugInInterface = 0;
	IOUSBInterfaceInterface190 **interface = 0;
	SInt32 score;
	HRESULT result = 0;
	
	err = (*device)->GetConfiguration(device, &configuration_id);
	if (configuration_id == 0) failExit("GetConfiguration", 0);
	if (err != KERN_SUCCESS) failExit("GetConfiguration", (int)err);
	
	request.bInterfaceClass = kUSBCommunicationDataInterfaceClass;
	request.bInterfaceSubClass = kIOUSBFindInterfaceDontCare;
	request.bInterfaceProtocol = kIOUSBFindInterfaceDontCare;
	request.bAlternateSetting = kIOUSBFindInterfaceDontCare;
	err = (*device)->CreateInterfaceIterator(device, &request, &iterator);
	if (err != KERN_SUCCESS) failExit("CreateInterfaceIterator", (int)err);
	
	usbInterface = IOIteratorNext(iterator);
	if (usbInterface == 0) failExit("Did not find USB interface", 0);

	if (iterator) IOObjectRelease(iterator);
	iterator = 0;
	
	err = IOCreatePlugInInterfaceForService(usbInterface, kIOUSBInterfaceUserClientTypeID, kIOCFPlugInInterfaceID, &plugInInterface, &score);
	if (err != KERN_SUCCESS) failExit("IOCreatePlugInInterfaceForService", (int)err);
	
	if (usbInterface) IOObjectRelease(usbInterface);
	usbInterface = 0;
	
	result = (*plugInInterface)->QueryInterface(plugInInterface, CFUUIDGetUUIDBytes (kIOUSBInterfaceInterfaceID183), (LPVOID)&interface);
	if (result) failExit("QueryInterface", (int)result);
	
	if (plugInInterface) (*plugInInterface)->Release(plugInInterface);
	plugInInterface = 0;
	
	return interface;
}

//-----------------------------------------------------------------------------------------------------------------------------------

usb_t getUSB() {
	kern_return_t err;
	mach_port_t masterPort = 0;
	IOUSBDeviceInterface** device;
	UInt8 numEndpoints, pipeRef;
	usb_t usb;

	err = IOMasterPort(MACH_PORT_NULL, &masterPort); //create a master port to talk to the IOKit
	if (err != KERN_SUCCESS) failExit("IOMasterPort", (int)err);
	device = findDevice(masterPort, ATMEL_VID, ATMEL_ARM_PID, USB_ADDR_ANY, USB_LOC_ANY);
	mach_port_deallocate(mach_task_self(), masterPort);  
	setConfiguration(device, USB_DEV_CONFIG);
	usb.interface = findInterface(device);
	(*device)->Release(device);

	//setup pipes
	err = (*(usb.interface))->USBInterfaceOpen(usb.interface);
	if (err != KERN_SUCCESS) failExit("USBInterfaceOpen", err);
	err = (*(usb.interface))->GetNumEndpoints(usb.interface, &numEndpoints);
	if (err != KERN_SUCCESS) failExit("GetNumEndpoints", err);
	for (pipeRef=1 ; pipeRef<=numEndpoints ; pipeRef++) {
		UInt8 direction, number, transferType, interval;
		UInt16 maxPacketSize;
		err = (*(usb.interface))->GetPipeProperties(usb.interface, pipeRef, &direction, &number, &transferType, &maxPacketSize, &interval);
		if (err != KERN_SUCCESS) failExit("GetPipeProperties", err);
		if (direction == kUSBIn) {
			usb.pipeIn = pipeRef;
		} else {
			usb.pipeOut = pipeRef;
		}
	}
	
	
	return usb;
}

void writeUSB (usb_t usb, char *buffer, UInt32 count) {
	kern_return_t err;
	
	err = (*(usb.interface))->WritePipe(usb.interface, usb.pipeOut, buffer, count);
	if (err != KERN_SUCCESS) failExit("WritePipe", (int)err);
}

//returns buffer size; buffer is null-terminated
int readUSB(usb_t usb, char* rslt, int size) {
	kern_return_t err;
	char buffer[128];
	int count = 0;
	UInt32 n, i;
	
	for(;;) {
		n = sizeof(buffer);
		err = (*(usb.interface))->ReadPipeTO(usb.interface, usb.pipeIn, buffer, &n, 200, 0);
		switch (err) {
			case KERN_SUCCESS:
				if (count + n >= size) failExit("ReadPipe buffer overrun", 0);
				for (i=0 ; i<n ; i++) {
					*rslt = buffer[i];
					++rslt;
					++count;
					if (buffer[i] == SAMBA_PROMPT) {
						*rslt = '\0';
						++count;
						return count;
					}
				}
				break;
			case kIOUSBTransactionTimeout:
				printf("Read timeout\n");
				(*(usb.interface))->ClearPipeStallBothEnds(usb.interface, usb.pipeIn);
				break;
			default:
				failExit("ReadPipe", (int)err);
		}
	}
}

void readCountUSB(usb_t usb, int count) {
	kern_return_t err;
	char buffer[128];
	UInt32 n;
	int c = 0;
	
	for(;;) {
		n = sizeof(buffer);
		err = (*(usb.interface))->ReadPipeTO(usb.interface, usb.pipeIn, buffer, &n, 200, 0);
		switch (err) {
			case KERN_SUCCESS:
				c += n;
				if (c >= count) return;
				break;
			case kIOUSBTransactionTimeout:
				printf("Read timeout\n");
				(*(usb.interface))->ClearPipeStallBothEnds(usb.interface, usb.pipeIn);
				break;
			default:
				failExit("ReadPipe", (int)err);
		}
	}

}
